/*******************************************************************************
  MPLAB Harmony Application Source File

  Company:
    Microchip Technology Inc.

  File Name:
    app.c

  Summary:
    This file contains the source code for the MPLAB Harmony application.

  Description:
    This file contains the source code for the MPLAB Harmony application.  It
    implements the logic of the application's state machine and it may call
    API routines of other MPLAB Harmony modules in the system, such as drivers,
    system services, and middleware.  However, it does not call any of the
    system interfaces (such as the "Initialize" and "Tasks" functions) of any of
    the modules in the system or make any assumptions about when those functions
    are called.  That is the responsibility of the configuration-specific system
    files.
 *******************************************************************************/

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#include "app.h"
#include "cryptoauthlib.h"
#include "host/atca_host.h"

#include "atcacert/atcacert_client.h"
#include "atcacert/atcacert_def.h"
#include "atcacert/atcacert_host_hw.h"
#include "atcacert/atcacert_pem.h"

#include "../../../project_config.h"

// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
#define ACCESSORY_PRIVKEY_SLOT      (0x00)

extern ATCAIfaceCfg atecc608_0_init_data;
extern ATCAIfaceCfg ecc204_0_init_data;
extern ATCAIfaceCfg ecc204_1_init_data;

extern const uint8_t g_cert_ca_public_key_1_signer[ATCA_PUB_KEY_SIZE];
extern const atcacert_def_t g_cert_def_1_signer;
extern const atcacert_def_t g_cert_def_3_device;

char displaystr[200];
size_t displaylen;

uint8_t device_public_key[ATCA_PUB_KEY_SIZE];
const char json_message[] = JSON_MESSAGE_STRING;
uint8_t message_digest[SHA_RSP_SIZE];
uint8_t signature[SIGN_RSP_SIZE];

ATCAIfaceCfg *ecc204_cfg = (SELECTED_ECC204_SWI) ? &ecc204_1_init_data : &ecc204_0_init_data;

// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    This structure should be initialized by the APP_Initialize function.

    Application strings and buffers are be defined outside this structure.
*/
APP_DATA appData;

ATCA_STATUS rebuild_and_verify_certs(void)
{
    ATCA_STATUS status;
    uint8_t signer_cert[1024];
    size_t  signer_cert_size = 1024;
    uint8_t device_cert[1024];
    size_t  device_cert_size = 1024;
    
    uint8_t signer_public_key[ATCA_PUB_KEY_SIZE];
    
    //Rebuild certificates
    do
    {
        signer_cert_size = sizeof(signer_cert);
        if (ATCACERT_E_SUCCESS != (status = atcacert_read_cert(&g_cert_def_1_signer, g_cert_ca_public_key_1_signer,
                                                            signer_cert, &signer_cert_size)))
        {
            printf("\r\natcacert_read_cert failed with an error %d\r\n", status);
            break;
        }

        if (ATCACERT_E_SUCCESS != (status = atcacert_get_subj_public_key(&g_cert_def_1_signer, signer_cert,
                                                                      signer_cert_size, signer_public_key)))
        {
            printf("\r\natcacert_get_subj_public_key failed with an error %d\r\n", status);
            break;
        }

        device_cert_size = sizeof(device_cert);
        if (ATCACERT_E_SUCCESS != (status = atcacert_read_cert(&g_cert_def_3_device, signer_public_key,
                                                            device_cert, &device_cert_size)))
        {
            printf("\r\natcacert_read_cert failed with an error %d\r\n", status);
            break;
        }
    }
    while(0);
    
    atcab_release();
    
    //Connect to host 
    if (ATCA_SUCCESS != (status = (int)atcab_init(&atecc608_0_init_data)))
    {
        printf("\r\natcab_init(ECC608 I2C) failed with status {%02X}.\r\n", status);
        return status;
    }
    //Verify cert on host
    if (ATCACERT_E_SUCCESS != (status = atcacert_verify_cert_hw(&g_cert_def_1_signer, signer_cert,
                                                             signer_cert_size, g_cert_ca_public_key_1_signer)))
    {
        printf("\r\nSigner certificate is not verified with an error as %d\r\n", status);

        if (ATCACERT_E_SUCCESS != (status = atcacert_verify_cert_hw(&g_cert_def_3_device, device_cert,
                                                                 device_cert_size, signer_public_key)))
        {
            printf("\r\nDevice certificate is not verified with an error as %d\r\n", status);
        }
    }

    if (ATCACERT_E_SUCCESS != (status = atcacert_get_subj_public_key(&g_cert_def_3_device, device_cert,
                                                                 device_cert_size, device_public_key)))
    {
        printf("\r\natcacert_get_subj_public_key failed with an error %d\r\n", status);
    }

    printf("success\r\n");
    printf("Signer cert validated.\r\nDevice cert validated.\r\n");
    return status;
}

ATCA_STATUS authenticate_meter(void)
{
    ATCA_STATUS status;
    uint8_t challenge[RANDOM_NUM_SIZE];
    uint8_t response[ATCA_SIG_SIZE];

    //Generate auth challenge
    if (ATCACERT_E_SUCCESS != (status = atcacert_gen_challenge_hw(challenge)))
    {
        printf("\r\nHost generate challenge failed with an error %d\r\n", status);
        return status;
    }
    atcab_release();
    
    //Get response from device
    if (ATCA_SUCCESS != (status = atcab_init(ecc204_cfg)))
    {
        printf("\r\natcab_init(ECC204 I2C) failed with status {%02X}.\r\n", status);
        return status;
    }
    if (ATCACERT_E_SUCCESS != (status = atcacert_get_response(g_cert_def_3_device.private_key_slot,
                                                           challenge, response)))
    {
        printf("\r\natcacert_get_response failed with an error %d\r\n", status);
    }
    atcab_release();
    
    //Validate response on host
    if (ATCA_SUCCESS != (status = (int)atcab_init(&atecc608_0_init_data)))
    {
        printf("\r\natcab_init(ECC608 I2C) failed with status {%02X}.\r\n", status);
        return status;
    }
    if (ATCACERT_E_SUCCESS != (status = atcacert_verify_response_hw(device_public_key, challenge, response)))
    {
        printf("\r\natcacert_verify_response_hw failed with an error %d\r\n", status);
        return status;
    }
    
    printf("success\r\n");
    printf("Challenge passed.\r\nDevice authenticated successfully.\r\n");
    return status;
}

ATCA_STATUS generate_json_msg_digest(void)
{
    ATCA_STATUS status;
    
    //Perform SHA256 Hash of JSON message
    if (ATCA_SUCCESS == (status = atcab_sha(strlen(json_message), (const uint8_t*) json_message, message_digest)))
    {
        printf("success\r\n");
        printf("JSON Message cleartext:\r\n%s\r\n", json_message);
        displaylen = sizeof(displaystr);
        atcab_bin2hex(message_digest, ATCA_SHA256_DIGEST_SIZE, displaystr, &displaylen);
        printf("Message digest:\r\n%s\r\n", displaystr);
    }
    else
    {
        printf("failed, status: {%02X}", status);
    }
    return status;
}

ATCA_STATUS sign_json_message(void)
{
    ATCA_STATUS status;
    //Get signature of message digest from accessory
    if (ATCA_SUCCESS == (status = atcab_sign(ACCESSORY_PRIVKEY_SLOT, message_digest, signature)))
    {
        printf("success\r\n");
        displaylen = sizeof(displaystr);
        atcab_bin2hex(signature, SIGN_RSP_SIZE, displaystr, &displaylen);
        printf("Message signature:\r\n%s\r\n", displaystr);
    }
    else
    {
        printf("failed, status: {%02X}", status);
    }
    return status;
}

ATCA_STATUS host_verify_json_message(void)
{
    ATCA_STATUS status;
    
    //Perform SHA256 hash on host device
    if (ATCA_SUCCESS != (status = atcab_sha(strlen(json_message), (const uint8_t*) json_message, message_digest)))
    {
        printf("failed, status: {%02X}", status);
    }
    
    //Verify signature from the accessory on the host
    if (ATCA_SUCCESS == (status = atcab_verify(VERIFY_MODE_EXTERNAL, VERIFY_KEY_P256, signature, device_public_key, NULL, NULL)))
    {
        printf("success\r\nApplication code complete\r\n");
    }
    else
    {
        printf("failed, status: {%02X}", status);
    }
    return status;
}


// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Initialize(void)
{
    /* Place the App state machine in its initial state. */
    appData.state = APP_STATE_INIT;

    /* TODO: Initialize your application's state machine and other
     * parameters.
     */
}

/******************************************************************************
  Function:
    void APP_Tasks ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Tasks(void)
{
    ATCA_STATUS status;

    /* Check the application's current state. */
    switch (appData.state)
    {
        /* Application's initial state. */
        case APP_STATE_INIT:
            if (ATCA_SUCCESS != (status = atcab_init(ecc204_cfg)))
            {
                printf("\natcab_init(ECC204 I2C) failed with status {%02X}.\r\n", status);
            }
            appData.state = (status == ATCA_SUCCESS) ? APP_REBUILD_AND_VERIFY_CERTS : APP_STATE_FINISH;
            break;
            
        case APP_REBUILD_AND_VERIFY_CERTS:
            printf("Rebuilding certs and verifying...          ");
            status = rebuild_and_verify_certs();
            appData.state = (status == ATCACERT_E_SUCCESS) ? APP_STATE_AUTHENTICATE_METER : APP_STATE_FINISH;
            
        case APP_STATE_AUTHENTICATE_METER:
            printf("Authenticating meter...                    ");
            status = authenticate_meter();
            appData.state = (status == ATCACERT_E_SUCCESS) ? APP_STATE_GET_MESSAGE_HASH : APP_STATE_FINISH;

        case APP_STATE_GET_MESSAGE_HASH:
            if (ATCA_SUCCESS != (status = atcab_init(ecc204_cfg)))
            {
                printf("\natcab_init(ECC204 I2C) failed with status {%02X}.\r\n", status);
            }
            printf("Performing SHA256 hash of JSON message...  ");
            appData.state = (generate_json_msg_digest() == ATCA_SUCCESS) ? APP_STATE_SIGN_MESSAGE_DIGEST : APP_STATE_FINISH;
            break;

        case APP_STATE_SIGN_MESSAGE_DIGEST:
            printf("Getting ECDSA signature of the message...  ");
            appData.state = (sign_json_message() == ATCA_SUCCESS) ? APP_STATE_HOST_VERIFY_MESSAGE : APP_STATE_FINISH;
            break;

        case APP_STATE_HOST_VERIFY_MESSAGE:
            if (ATCA_SUCCESS == (status = (int)atcab_init(&atecc608_0_init_data)))
            {
                printf("Verifying message signature on host...     ");
                status = host_verify_json_message();
            }
            else
            {
                printf("\natcab_init(ECC608 I2C) failed with status {%02X}.\r\n", status);
            }
            appData.state = APP_STATE_FINISH;
            break;

        case APP_STATE_FINISH:
            break;

        /* The default state should never be executed. */
        default:
        {
            /* TODO: Handle error in application's state machine. */
            break;
        }
    }
}

/*******************************************************************************
 End of File
 */
